#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################


from pyomo.environ import (
    NonNegativeReals,
    Param,
    Var,
    check_optimal_termination,
    exp,
    units as pyunits,
)

from idaes.core import UnitModelBlockData
from idaes.core.solvers import get_solver
from idaes.core.util import scaling as iscale
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.misc import add_object_reference
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog
from watertap.core import InitializationMixin

from .MD_channel_base import (
    ConcentrationPolarizationType,
    TemperaturePolarizationType,
    MassTransferCoefficient,
)

__author__ = "Elmira Shamlou"


class MembraneDistillationBaseData(InitializationMixin, UnitModelBlockData):
    def build(self):
        """
        Common variables and constraints for a DCMD unit model

        """

        super().build()

        self._make_MD_channel_control_volume("hot_ch", self.config.hot_ch)

        self.hot_ch.add_state_blocks(
            has_phase_equilibrium=False,
            flow_direction=self.config.hot_ch.flow_direction,
            property_package_vapor=self.config.hot_ch.property_package_vapor,
            property_package_args_vapor=self.config.hot_ch.property_package_args_vapor,
        )

        # self.hot_ch.set_config(self.config.hot_ch)

        self.hot_ch.add_material_balances(
            balance_type=self.config.hot_ch.material_balance_type,
            has_mass_transfer=True,
        )

        self.hot_ch.add_momentum_balances(
            balance_type=self.config.hot_ch.momentum_balance_type,
            pressure_change_type=self.config.hot_ch.pressure_change_type,
            has_pressure_change=self.config.hot_ch.has_pressure_change,
            friction_factor=self.config.hot_ch.friction_factor,
        )

        self.hot_ch.add_energy_balances(
            balance_type=self.config.hot_ch.energy_balance_type,
            has_heat_transfer=True,
            has_enthalpy_transfer=True,
        )

        # Add constraint for volumetric flow equality between interface and bulk
        self.hot_ch.add_extensive_flow_to_interface()

        self.hot_ch.add_concentration_polarization(
            concentration_polarization_type=self.config.hot_ch.concentration_polarization_type,
            mass_transfer_coefficient=self.config.hot_ch.mass_transfer_coefficient,
        )

        self.hot_ch.add_temperature_polarization(
            temperature_polarization_type=self.config.hot_ch.temperature_polarization_type,
        )

        self._make_MD_channel_control_volume("cold_ch", self.config.cold_ch)

        self.cold_ch.add_state_blocks(
            has_phase_equilibrium=False,
            flow_direction=self.config.cold_ch.flow_direction,
            property_package_vapor=self.config.cold_ch.property_package_vapor,
            property_package_args_vapor=self.config.cold_ch.property_package_args_vapor,
        )

        # self.cold_ch.set_config(self.config.cold_ch)

        self.cold_ch.add_material_balances(
            balance_type=self.config.cold_ch.material_balance_type,
            has_mass_transfer=True,
        )

        self.cold_ch.add_momentum_balances(
            balance_type=self.config.cold_ch.momentum_balance_type,
            pressure_change_type=self.config.cold_ch.pressure_change_type,
            has_pressure_change=self.config.cold_ch.has_pressure_change,
            friction_factor=self.config.cold_ch.friction_factor,
        )

        self.cold_ch.add_energy_balances(
            balance_type=self.config.cold_ch.energy_balance_type,
            has_heat_transfer=True,
            has_enthalpy_transfer=True,
        )

        # Add constraint for volumetric flow equality between interface and bulk
        self.cold_ch.add_extensive_flow_to_interface()

        # Concentration polarization constraint is not accounted for in the below method; it is
        # written later in the base model (eq_concentration_polarization)
        self.cold_ch.add_concentration_polarization(
            concentration_polarization_type=ConcentrationPolarizationType.none,
            mass_transfer_coefficient=MassTransferCoefficient.none,
        )

        self.cold_ch.add_temperature_polarization(
            temperature_polarization_type=self.config.cold_ch.temperature_polarization_type,
        )
        add_object_reference(self, "length_domain", self.hot_ch.length_domain)
        add_object_reference(
            self, "difference_elements", self.hot_ch.difference_elements
        )
        add_object_reference(self, "first_element", self.hot_ch.first_element)
        add_object_reference(self, "nfe", self.hot_ch.nfe)

        # Add Ports
        self.add_inlet_port(name="hot_ch_inlet", block=self.hot_ch)
        self.add_outlet_port(name="hot_ch_outlet", block=self.hot_ch)
        self.add_inlet_port(name="cold_ch_inlet", block=self.cold_ch)
        self.add_outlet_port(name="cold_ch_outlet", block=self.cold_ch)

        self._add_heat_flux()
        self._add_mass_flux()

        self.recovery_mass = Var(
            self.flowsheet().config.time,
            # self.config.hot_ch.property_package.phase_list,
            # self.config.hot_ch.property_package.component_list,
            initialize=0.1,
            bounds=(1e-11, 0.99),
            units=pyunits.dimensionless,
            doc="Mass-based recovery",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            # self.config.hot_ch.property_package.component_list,
        )
        def eq_recovery_mass(b, t):
            return (
                b.recovery_mass[t]
                * b.hot_ch.properties[t, b.first_element].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                == b.cold_ch.properties[t, b.first_element].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                - b.cold_ch.properties[
                    t, b.cold_ch.length_domain.last()
                ].flow_mass_phase_comp["Liq", "H2O"]
            )

        self._add_mass_transfer()
        self._add_heat_transfer()

        @self.Expression(
            self.flowsheet().config.time,
            doc="Average thermal efficiency",
        )
        def thermal_efficiency(b, t):
            total_enth_flux = sum(
                b.flux_enth_hot[t, x] for x in self.difference_elements
            )
            total_cond_heat_flux = sum(
                b.flux_conduction_heat[t, x] for x in self.difference_elements
            )

            return total_enth_flux / (total_enth_flux + total_cond_heat_flux)

        @self.Expression(
            self.flowsheet().config.time,
            doc="module heat recovery: ratio of the actual heat recovered by cold side to the maximum ideal heat recovery",
        )
        def effectiveness(b, t):

            return (
                b.cold_ch.properties[t, b.first_element].temperature
                - b.cold_ch.properties[t, b.cold_ch.length_domain.last()].temperature
            ) / (
                b.hot_ch.properties[t, b.first_element].temperature
                - b.cold_ch.properties[t, b.cold_ch.length_domain.last()].temperature
            )

    def _add_mass_flux(self):

        solute_set = self.config.hot_ch.property_package.solute_set

        units_meta = (
            self.config.hot_ch.property_package.get_metadata().get_derived_units
        )

        # todo: add calculation method for permeability coefficient
        self.permeability_coef = Var(
            self.flowsheet().config.time,
            initialize=1e-10,
            bounds=(1e-11, 1e-9),
            units=units_meta("mass")
            * units_meta("time") ** -1
            * units_meta("length") ** -1
            * units_meta("pressure") ** -1,
            doc="membrane permeabiity coeffcient",
        )

        self.flux_mass = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            initialize=1e-10,
            bounds=(1e-12, 1e-1),
            units=units_meta("mass")
            * units_meta("time") ** -1
            * units_meta("length") ** -2,
            doc="solvent vapor mass flux across the membrane",
        )

        self.flux_enth_hot = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            initialize=1e3,
            bounds=(1e-10, 1e5),
            units=pyunits.J * pyunits.s**-1 * pyunits.m**-2,
            doc="hot side evaporation enthalpy flux",
        )

        self.flux_enth_cold = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            initialize=1e3,
            bounds=(1e-10, 1e5),
            units=pyunits.J * pyunits.s**-1 * pyunits.m**-2,
            doc="cold side condensation enthalpy flux",
        )

        self.dens_solvent = Param(
            initialize=1000,
            units=units_meta("mass") * units_meta("length") ** -3,
            doc="Pure water density",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Solvent mass flux",
        )
        def eq_flux_mass(b, t, x):
            return b.flux_mass[t, x] == b.permeability_coef[
                t
            ] / b.membrane_thickness * (
                b.hot_ch.properties_interface[t, x].pressure_sat
                - b.cold_ch.properties_interface[t, x].pressure_sat
            )

        @self.Expression(
            self.flowsheet().config.time,
            doc="Average mass flux expression",
        )
        def flux_mass_avg(b, t):
            return sum(b.flux_mass[t, x] for x in self.difference_elements) / self.nfe

        if (
            self.config.hot_ch.concentration_polarization_type
            == ConcentrationPolarizationType.calculated
        ):

            @self.Constraint(
                self.flowsheet().config.time,
                self.difference_elements,
                solute_set,
                doc="External Concentration polarization in hot side",
            )
            def eq_concentration_polarization_hot_ch(b, t, x, j):
                jw = b.flux_mass[t, x] / self.dens_solvent
                exponent = jw / self.hot_ch.K[t, x, j]
                return b.hot_ch.properties_interface[t, x].conc_mass_phase_comp[
                    "Liq", j
                ] == b.hot_ch.properties[t, x].conc_mass_phase_comp["Liq", j] * exp(
                    exponent
                )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="hot side evaporation enthalpy flux",
        )
        def eq_flux_hot_enth(b, t, x):
            return (
                b.flux_enth_hot[t, x]
                == b.flux_mass[t, x]
                * b.hot_ch.properties_vapor[t, x].enth_mass_phase["Vap"]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="cold side evaporation enthalpy flux",
        )
        def eq_flux_cold_enth(b, t, x):
            return (
                b.flux_enth_cold[t, x]
                == b.flux_mass[t, x]
                * b.cold_ch.properties_vapor[t, x].enth_mass_phase["Vap"]
            )

        @self.Expression(
            self.flowsheet().config.time,
            doc="Average hot side enthalpy flux expression",
        )
        def flux_enth_hot_avg(b, t):
            return (
                sum(b.flux_enth_hot[t, x] for x in self.difference_elements) / self.nfe
            )

        @self.Expression(
            self.flowsheet().config.time,
            doc="Average cold side enthalpy flux expression",
        )
        def flux_enth_cold_avg(b, t):
            return (
                sum(b.flux_enth_cold[t, x] for x in self.difference_elements) / self.nfe
            )

        @self.Constraint(
            self.flowsheet().time,
            self.difference_elements,
            doc="hot side Vapor temperature",
        )
        def eq_vapor_temperature_hot(b, t, x):
            return (
                b.hot_ch.properties_vapor[t, x].temperature
                == b.hot_ch.properties_interface[t, x].temperature
            )

        @self.Constraint(
            self.flowsheet().time,
            self.difference_elements,
            doc="hot  side Vapor pressure",
        )
        def eq_vapor_pressure_hot(b, t, x):
            return (
                b.hot_ch.properties_vapor[t, x].pressure
                == b.hot_ch.properties_interface[t, x].pressure_sat
            )

        @self.Constraint(
            self.flowsheet().time,
            self.difference_elements,
            doc="cold side Vapor temperature",
        )
        def eq_vapor_temperature_cold(b, t, x):
            return (
                b.cold_ch.properties_vapor[t, x].temperature
                == b.cold_ch.properties_interface[t, x].temperature
            )

        @self.Constraint(
            self.flowsheet().time,
            self.difference_elements,
            doc="cold  side Vapor pressure",
        )
        def eq_vapor_pressure_cold(b, t, x):
            return (
                b.cold_ch.properties_vapor[t, x].pressure
                == b.cold_ch.properties_interface[t, x].pressure_sat
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Vapor flow rate and no liquid in vapor state block",
        )
        def eq_vapor_flow(b, t, x):
            lb = b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Liq", "H2O"].lb
            b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Liq", "H2O"].fix(lb)

            return (
                b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Vap", "H2O"]
                == b.flux_mass[t, x] * b.area / self.nfe
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Equal vapor flow rate between hot_ch and cold_ch",
        )
        def eq_vapor_flow_equal(b, t, x):
            lb = b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Liq", "H2O"].lb
            b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Liq", "H2O"].fix(lb)
            return (
                b.hot_ch.properties_vapor[t, x].flow_mass_phase_comp["Vap", "H2O"]
                == b.cold_ch.properties_vapor[t, x].flow_mass_phase_comp["Vap", "H2O"]
            )

        # Check for hot channel temperature polarization type
        if (
            self.config.hot_ch.temperature_polarization_type
            != TemperaturePolarizationType.none
        ):

            @self.Constraint(
                self.flowsheet().config.time,
                self.difference_elements,
                doc="Temperature polarization in hot channel",
            )
            def eq_temperature_polarization_hot(b, t, x):
                return (
                    b.hot_ch.h_conv[t, x]
                    * (
                        b.hot_ch.properties[t, x].temperature
                        - b.hot_ch.properties_interface[t, x].temperature
                    )
                    == b.flux_conduction_heat[t, x]
                    + b.flux_enth_hot[t, x]
                    - b.flux_mass[t, x]
                    * b.hot_ch.properties[t, x].enth_mass_phase["Liq"]
                )

        # Check for cold channel temperature polarization type
        if (
            self.config.cold_ch.temperature_polarization_type
            != TemperaturePolarizationType.none
        ):

            @self.Constraint(
                self.flowsheet().config.time,
                self.difference_elements,
                doc="Temperature polarization in cold channel",
            )
            def eq_temperature_polarization_cold(b, t, x):
                return (
                    b.cold_ch.h_conv[t, x]
                    * (
                        -b.cold_ch.properties[t, x].temperature
                        + b.cold_ch.properties_interface[t, x].temperature
                    )
                    == b.flux_conduction_heat[t, x]
                    + b.flux_enth_cold[t, x]
                    - b.flux_mass[t, x]
                    * b.cold_ch.properties[t, x].enth_mass_phase["Liq"]
                )

        return self.eq_flux_mass

    def _add_heat_flux(self):

        # todo: add calculation method for permeability coefficient
        self.membrane_thickness = Var(
            initialize=1e-4,
            bounds=(1e-5, 1e-2),
            doc="membrane thickness",
            units=pyunits.m,
        )

        self.membrane_tc = Var(
            initialize=0.2,
            bounds=(0, 1),
            units=pyunits.J * pyunits.s**-1 * pyunits.K**-1 * pyunits.m**-1,
            doc="Thermal conductivity coefficient of the membrane",
        )

        self.flux_conduction_heat = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            initialize=1e-4,
            bounds=(0, 1e10),
            units=pyunits.J * pyunits.s**-1 * pyunits.m**-2,
            doc="conduction heat flux",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="conduction heat flux",
        )
        def eq_flux_heat(b, t, x):
            return b.flux_conduction_heat[
                t, x
            ] == b.membrane_tc / b.membrane_thickness * (
                b.hot_ch.properties_interface[t, x].temperature
                - b.cold_ch.properties_interface[t, x].temperature
            )

        @self.Expression(
            self.flowsheet().config.time,
            doc="Average conduction heat flux expression",
        )
        def flux_conduction_heat_avg(b, t):
            return (
                sum(b.flux_conduction_heat[t, x] for x in self.difference_elements)
                / self.nfe
            )

        return self.eq_flux_heat

    def _add_length_and_width(self):
        units_meta = (
            self.config.hot_ch.property_package.get_metadata().get_derived_units
        )
        self.length = Var(
            initialize=10,
            bounds=(0.1, 5e2),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="Effective membrane length",
        )
        self.width = Var(
            initialize=1,
            bounds=(1e-1, 1e3),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="Membrane width",
        )

    def _add_area(self, include_constraint=True):

        if not hasattr(self, "area"):
            self.area = Var(
                initialize=10,
                bounds=(1e-1, 1e3),
                domain=NonNegativeReals,
                units=pyunits.m**2,
                doc="Total Membrane area",
            )

        if include_constraint:
            if not hasattr(self, "eq_area"):
                # Membrane area equation
                @self.Constraint(doc="Total Membrane area")
                def eq_area(b):
                    return b.area == b.length * b.width

            else:
                raise ValueError(
                    "include_constraint was set to True inside of _add_area(), but area constraint already exists."
                )

    def initialize_build(
        self,
        initialize_guess=None,
        state_args_hot_ch=None,
        state_args_cold_ch=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        hot_ch_flags = self.hot_ch.initialize(
            state_args=state_args_hot_ch,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            initialize_guess=initialize_guess,
            type="hot_ch",
        )

        init_log.info_high("Initialization Step 1a (hot channel) Complete")

        cold_ch_flags = self.cold_ch.initialize(
            state_args=state_args_cold_ch,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            initialize_guess=initialize_guess,
            type="cold_ch",
        )

        init_log.info_high("Initialization Step 1b (cold channel) Complete")

        if degrees_of_freedom(self) != 0:
            raise Exception(
                f"{self.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(self)}"
            )

        # solver
        opt = get_solver(solver, optarg)

        # Solve unit *without* any flux equations
        self.eq_flux_mass.deactivate()
        self.eq_flux_heat.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high(f"Initialization Step 2 (No Flux) {idaeslog.condition(res)}")

        # Activate only the heat flux equations
        self.eq_flux_heat.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high(
            f"Initialization Step 3 (Heat Flux Only) {idaeslog.condition(res)}"
        )

        # Activate mass flux equations as well
        self.eq_flux_mass.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high(
            f"Initialization Step 4 (Heat and Mass Flux) {idaeslog.condition(res)}"
        )

        # Release inlet state
        self.cold_ch.release_state(cold_ch_flags, outlvl)
        self.hot_ch.release_state(hot_ch_flags, outlvl)

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot channel Inlet": self.hot_ch_inlet,
                "Hot channel Outlet": self.hot_ch_outlet,
                "Cold channel Inlet": self.cold_ch_inlet,
                "Cold channel Outlet": self.cold_ch_outlet,
            },
            time_point=time_point,
        )

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        expr_dict = {}

        var_dict["Mass Recovery Rate"] = self.recovery_mass[time_point]
        var_dict["Membrane Area"] = self.area
        if hasattr(self, "length"):
            var_dict["Membrane Length"] = self.length
        if hasattr(self, "width"):
            var_dict["Membrane Width"] = self.width

        expr_dict["Average Solute Flux (LMH)"] = self.flux_mass_avg[time_point] * 1000
        expr_dict["Thermal efficiency (%)"] = self.thermal_efficiency[time_point] * 100
        expr_dict["Effectiveness (%)"] = self.effectiveness[time_point] * 100

        return {"vars": var_dict, "exprs": expr_dict}

    def calculate_scaling_factors(self):

        if iscale.get_scaling_factor(self.dens_solvent) is None:
            sf = iscale.get_scaling_factor(
                self.hot_ch.properties[0, self.first_element].dens_mass_phase["Liq"]
            )
            iscale.set_scaling_factor(self.dens_solvent, sf)

        super().calculate_scaling_factors()
        for t, v in self.recovery_mass.items():
            sf = 1
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        if iscale.get_scaling_factor(self.area) is None:
            sf = iscale.get_scaling_factor(self.area, default=10, warning=True)
            iscale.set_scaling_factor(self.area, sf)

        if iscale.get_scaling_factor(self.permeability_coef) is None:
            iscale.set_scaling_factor(self.permeability_coef, 1e10)

        if iscale.get_scaling_factor(self.membrane_thickness) is None:
            iscale.set_scaling_factor(self.membrane_thickness, 1e4)

        if iscale.get_scaling_factor(self.membrane_tc) is None:
            iscale.set_scaling_factor(self.membrane_tc, 10)

        for (t, x), v in self.flux_mass.items():
            if iscale.get_scaling_factor(v) is None:

                sf_pressure_hot = iscale.get_scaling_factor(
                    self.hot_ch.properties_interface[t, x].pressure_sat
                )

                sf_permeability = iscale.get_scaling_factor(self.permeability_coef)
                sf_thickness = iscale.get_scaling_factor(self.membrane_thickness)

                sf_flux = sf_permeability * (sf_pressure_hot) / sf_thickness

                iscale.set_scaling_factor(v, sf_flux)

                iscale.constraint_scaling_transform(self.eq_flux_mass[t, x], sf_flux)

        for (t, x), v in self.flux_enth_hot.items():
            if iscale.get_scaling_factor(v) is None:
                sf_flux_enth = sf_flux * iscale.get_scaling_factor(
                    self.hot_ch.properties_vapor[t, x].enth_mass_phase["Vap"]
                )
                iscale.set_scaling_factor(v, sf_flux_enth)

        for (t, x), v in self.flux_enth_cold.items():
            if iscale.get_scaling_factor(v) is None:
                sf_flux_enth = sf_flux * iscale.get_scaling_factor(
                    self.cold_ch.properties_vapor[t, x].enth_mass_phase["Vap"]
                )
                iscale.set_scaling_factor(v, sf_flux_enth)

        for (t, x), v in self.flux_conduction_heat.items():
            if iscale.get_scaling_factor(v) is None:
                sf_flux_cond = (
                    iscale.get_scaling_factor(self.membrane_tc)
                    / iscale.get_scaling_factor(self.membrane_thickness)
                    * iscale.get_scaling_factor(
                        self.hot_ch.properties_interface[t, x].temperature
                    )
                )
                iscale.set_scaling_factor(v, sf_flux_cond)

        if hasattr(self, "length"):
            if iscale.get_scaling_factor(self.length) is None:
                iscale.set_scaling_factor(self.length, 1)

        if hasattr(self, "width"):
            if iscale.get_scaling_factor(self.width) is None:
                iscale.set_scaling_factor(self.width, 1)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from watertap.tools.parameter_sweep.parameter_sweep import (
    ParameterSweep,
    RecursiveParameterSweep,
)
from watertap.tools.parameter_sweep.parameter_sweep_differential import (
    DifferentialParameterSweep,
)


def parameter_sweep(
    build_model,
    build_sweep_params,
    build_outputs=None,
    csv_results_file_name=None,
    h5_results_file_name=None,
    h5_parent_group_name=None,
    optimize_function=None,
    optimize_kwargs=None,
    reinitialize_function=None,
    reinitialize_kwargs=None,
    reinitialize_before_sweep=False,
    probe_function=None,
    debugging_data_dir=None,
    interpolate_nan_outputs=False,
    num_samples=None,
    seed=None,
    number_of_subprocesses=None,
    build_model_kwargs=None,
    build_sweep_params_kwargs=None,
    build_outputs_kwargs=None,
):
    """
    This function offers a general way to perform repeated optimizations
    of a model for the purposes of exploring a parameter space while
    monitoring multiple outputs.
    If provided, writes single CSV file to ``results_file`` with all inputs and resulting outputs.

    Arguments:

        build_model : A function that can be called to build a Pyomo ConcreteModel, OR
                      (deprecated) a Pyomo ConcreteModel containing a watertap flowsheet.

        build_sweep_params: A function that can be called to build a dictionary containing the values to vary
                            with the format ``sweep_params['Short/Pretty-print Name'] =
                            (model.fs.variable_or_param[index], lower_limit, upper_limit, num_samples)``.
                            A uniform number of samples ``num_samples`` will be take between
                            the ``lower_limit`` and ``upper_limit``.
                            OR (deprecated) the dictionary itself rather than a function for creating it.

        outputs (optional) : An optional function to produce a dictionary containing "short names" as
                             keys and and Pyomo objects on ``model`` whose values to report as values. E.g.,
                             ``outputs['Short/Pretty-print Name'] = model.fs.variable_or_expression_to_report``.
                             OR (deprecated) the dictionary itself rather than a function for creating it.
                             If neither is provided, i.e., outputs = None, the default behavior is to save all model
                             variables, parameters, and expressions which provides very thorough results
                             at the cost of large file sizes.

        csv_results_file_name (optional) : The path and file name to write a csv file. The default `None`
                                           does not write a csv file.

        h5_results_file_name (optional) : The path and file name to write a h5 file. The default `None`
                                          does not write a file.
                                          Writing an h5 file will also create a companion text file `{h5_results_file_name}.txt`
                                          which contains the variable names contained within the H5 file.

        h5_parent_group_name (optional) : Parent h5 groups for the parameter sweep inputs and outputs to be embedded in.
                                          The default is `None` and it accepts a string for the h5 group.

        optimize_function (optional) : A user-defined function to perform the optimization of flowsheet
                                       ``model`` and loads the results back into ``model``. The first
                                       argument of this function is ``model``. The default uses the
                                       default IDAES solver, raising an exception if the termination
                                       condition is not optimal.

        optimize_kwargs (optional) : Dictionary of kwargs to pass into every call to
                                     ``optimize_function``. The first arg will always be ``model``,
                                     e.g., ``optimize_function(model, **optimize_kwargs)``. The default
                                     uses no kwargs.

        reinitialize_function (optional) : A user-defined function to perform the re-initialize the
                                           flowsheet ``model`` if the first call to ``optimize_function``
                                           fails for any reason. After ``reinitialize_function``, the
                                           parameter sweep tool will immediately call
                                           ``optimize_function`` again.

        reinitialize_kwargs (optional) : Dictionary or kwargs to pass into every call to
                                         ``reinitialize_function``. The first arg will always be
                                         ``model``, e.g.,
                                         ``reinitialize_function(model, **reinitialize_kwargs)``.
                                         The default uses no kwargs.

        reinitialize_before_sweep (optional): Boolean option to reinitialize the flow sheet model before
                                              every parameter sweep realization. The default is False.
                                              Note the parameter sweep model will try to reinitialize the
                                              solve regardless of the option if the run fails.

        probe_function (optional): A user-defined function that can cheaply check if a current model
                                  configuration is solvable without actually reinitializing or solving.

        debugging_data_dir (optional) : Save results on a per-process basis for parallel debugging
                                        purposes. If None no `debugging` data will be saved.

        interpolate_nan_outputs (optional) : When the parameter sweep has finished, interior values
                                             of np.nan will be replaced with a value obtained via
                                             a linear interpolation of their surrounding valid neighbors.
                                             If true, a second output file with the extension "_clean"
                                             will be saved alongside the raw (un-interpolated) values.

        num_samples (optional) : If the user is using sampling techniques rather than a linear grid
                                 of values, they need to set the number of samples

        seed (optional) : If the user is using a random sampling technique, this sets the seed

        number_of_subprocesses (optional) : Directive for fanning out subprocesses to perform
                                            parallel computation.

        build_model_kwargs (optional): A dictionary of kwargs to pass into the build_model function.

        build_sweep_params_kwargs (optional): A dictionary of kwargs to pass into the build_sweep_params
                                              function.

    Returns:

        save_data : A list were the first N columns are the values of the parameters passed
                    by ``sweep_params`` and the remaining columns are the values of the
                    simulation identified by the ``outputs`` argument.
    """

    kwargs = {}
    if csv_results_file_name is not None:
        kwargs["csv_results_file_name"] = csv_results_file_name
    if h5_results_file_name is not None:
        kwargs["h5_results_file_name"] = h5_results_file_name
    if h5_parent_group_name is not None:
        kwargs["h5_parent_group_name"] = h5_parent_group_name
    if optimize_function is not None:
        kwargs["optimize_function"] = optimize_function
    if optimize_kwargs is not None:
        kwargs["optimize_kwargs"] = optimize_kwargs
    if reinitialize_function is not None:
        kwargs["reinitialize_function"] = reinitialize_function
    if reinitialize_kwargs is not None:
        kwargs["reinitialize_kwargs"] = reinitialize_kwargs
    kwargs["reinitialize_before_sweep"] = reinitialize_before_sweep
    if probe_function is not None:
        kwargs["probe_function"] = probe_function
    if debugging_data_dir is not None:
        kwargs["debugging_data_dir"] = debugging_data_dir
    if interpolate_nan_outputs is not None:
        kwargs["interpolate_nan_outputs"] = interpolate_nan_outputs
    if number_of_subprocesses is not None:
        kwargs["number_of_subprocesses"] = number_of_subprocesses
    if build_outputs_kwargs is not None:
        kwargs["build_outputs_kwargs"] = build_outputs_kwargs

    ps = ParameterSweep(**kwargs)

    return ps.parameter_sweep(
        build_model,
        build_sweep_params,
        build_outputs=build_outputs,
        num_samples=num_samples,
        seed=seed,
        build_model_kwargs=build_model_kwargs,
        build_sweep_params_kwargs=build_sweep_params_kwargs,
    )


def recursive_parameter_sweep(
    build_model,
    build_sweep_params,
    build_outputs=None,
    csv_results_file_name=None,
    h5_results_file_name=None,
    h5_parent_group_name=None,
    optimize_function=None,
    optimize_kwargs=None,
    reinitialize_function=None,
    reinitialize_kwargs=None,
    reinitialize_before_sweep=False,
    probe_function=None,
    debugging_data_dir=None,
    interpolate_nan_outputs=False,
    num_samples=None,
    seed=None,
    number_of_subprocesses=None,
):
    """
    This function is similar to the `parameter_sweep` function for exploring the parameter space while guranteeing a required number of solves.
    If provided, writes single CSV file to ``results_file`` with all inputs and resulting outputs.

    Arguments:

        model : A Pyomo ConcreteModel containing a watertap flowsheet, for best
                results it should be initialized before being passed to this
                function.

        sweep_params: A dictionary containing the values to vary with the format
                      ``sweep_params['Short/Pretty-print Name'] =
                      (model.fs.variable_or_param[index], lower_limit, upper_limit, num_samples)``.
                      A uniform number of samples ``num_samples`` will be take between
                      the ``lower_limit`` and ``upper_limit``.

        outputs : An optional dictionary containing "short names" as keys and and Pyomo objects
                  on ``model`` whose values to report as values. E.g.,
                  ``outputs['Short/Pretty-print Name'] = model.fs.variable_or_expression_to_report``.
                  If not provided, i.e., outputs = None, the default behavior is to save all model
                  variables, parameters, and expressions which provides very thorough results
                  at the cost of large file sizes.

        csv_results_file_name (optional) : The path and file name to write a csv file. The default `None`
                                           does not write a csv file.

        h5_results_file_name (optional) : The path and file name to write a h5 file. The default `None`
                                          does not write a file.
                                          Writing an h5 file will also create a companion text file `{h5_results_file_name}.txt`
                                          which contains the variable names contained within the H5 file.

        h5_parent_group_name (optional) : Parent h5 groups for the parameter sweep inputs and outputs to be embedded in.
                                          The default is `None` and it accepts a string for the h5 group.

        optimize_function (optional) : A user-defined function to perform the optimization of flowsheet
                                       ``model`` and loads the results back into ``model``. The first
                                       argument of this function is ``model``. The default uses the
                                       default IDAES solver, raising an exception if the termination
                                       condition is not optimal.

        optimize_kwargs (optional) : Dictionary of kwargs to pass into every call to
                                     ``optimize_function``. The first arg will always be ``model``,
                                     e.g., ``optimize_function(model, **optimize_kwargs)``. The default
                                     uses no kwargs.

        reinitialize_function (optional) : A user-defined function to perform the re-initialize the
                                           flowsheet ``model`` if the first call to ``optimize_function``
                                           fails for any reason. After ``reinitialize_function``, the
                                           parameter sweep tool will immediately call
                                           ``optimize_function`` again.

        reinitialize_kwargs (optional) : Dictionary or kwargs to pass into every call to
                                         ``reinitialize_function``. The first arg will always be
                                         ``model``, e.g.,
                                         ``reinitialize_function(model, **reinitialize_kwargs)``.
                                         The default uses no kwargs.

        reinitialize_before_sweep (optional): Boolean option to reinitialize the flow sheet model before
                                              every parameter sweep realization. The default is False.
                                              Note the parameter sweep model will try to reinitialize the
                                              solve regardless of the option if the run fails.

        probe_function (optional): A user-defined function that can cheaply check if a current model
                                  configuration is solvable without actually reinitializing or solving.

        debugging_data_dir (optional) : Save results on a per-process basis for parallel debugging
                                        purposes. If None no `debugging` data will be saved.

        interpolate_nan_outputs (optional) : When the parameter sweep has finished, interior values
                                             of np.nan will be replaced with a value obtained via
                                             a linear interpolation of their surrounding valid neighbors.
                                             If true, a second output file with the extension "_clean"
                                             will be saved alongside the raw (un-interpolated) values.

        num_samples (optional) : If the user is using sampling techniques rather than a linear grid
                                 of values, they need to set the required number of samples. This is the
                                 guaranteed number of solves that the user requires.

        seed (optional) : If the user is using a random sampling technique, this sets the seed

    Returns:

        save_data : A list were the first N columns are the values of the parameters passed
                    by ``sweep_params`` and the remaining columns are the values of the
                    simulation identified by the ``outputs`` argument.
    """

    kwargs = {}
    if csv_results_file_name is not None:
        kwargs["csv_results_file_name"] = csv_results_file_name
    if h5_results_file_name is not None:
        kwargs["h5_results_file_name"] = h5_results_file_name
    if h5_parent_group_name is not None:
        kwargs["h5_parent_group_name"] = h5_parent_group_name
    if optimize_function is not None:
        kwargs["optimize_function"] = optimize_function
    if optimize_kwargs is not None:
        kwargs["optimize_kwargs"] = optimize_kwargs
    if reinitialize_function is not None:
        kwargs["reinitialize_function"] = reinitialize_function
    if reinitialize_kwargs is not None:
        kwargs["reinitialize_kwargs"] = reinitialize_kwargs
    kwargs["reinitialize_before_sweep"] = reinitialize_before_sweep
    if probe_function is not None:
        kwargs["probe_function"] = probe_function
    if debugging_data_dir is not None:
        kwargs["debugging_data_dir"] = debugging_data_dir
    if interpolate_nan_outputs is not None:
        kwargs["interpolate_nan_outputs"] = interpolate_nan_outputs
    if number_of_subprocesses is not None:
        kwargs["number_of_subprocesses"] = number_of_subprocesses

    rps = RecursiveParameterSweep(**kwargs)

    return rps.parameter_sweep(
        build_model,
        build_sweep_params,
        build_outputs=build_outputs,
        num_samples=num_samples,
        seed=seed,
    )


def differential_parameter_sweep(
    build_model,
    build_sweep_params,
    build_differential_sweep_specs,
    build_outputs=None,
    csv_results_file_name=None,
    h5_results_file_name=None,
    h5_parent_group_name=None,
    optimize_function=None,
    optimize_kwargs=None,
    initialize_function=None,
    initialize_kwargs=None,
    initialize_before_sweep=False,
    probe_function=None,
    debugging_data_dir=None,
    interpolate_nan_outputs=False,
    num_samples=None,
    num_diff_samples=1,
    seed=None,
    guarantee_solves=False,
):
    """
    This function is similar to the `parameter_sweep` function for exploring the parameter space while guranteeing a required number of solves.
    If provided, writes single CSV file to ``results_file`` with all inputs and resulting outputs.

    Arguments:

        model : A Pyomo ConcreteModel containing a watertap flowsheet, for best
                results it should be initialized before being passed to this
                function.

        sweep_params: A dictionary containing the values to vary with the format
                      ``sweep_params['Short/Pretty-print Name'] =
                      (model.fs.variable_or_param[index], lower_limit, upper_limit, num_samples)``.
                      A uniform number of samples ``num_samples`` will be take between
                      the ``lower_limit`` and ``upper_limit``.

        differential_sweep_specs: A specification dictionary that contains details for how to construct the parameter sweep dictionary for differential sweep.
            This is a nested dictionary where the first level denotes the variable names for which the differential sweep needs to be carried out.
            The second level denotes various options to be used for wach variable.
            The number of samples for each differential sweep is specified while initializing the DifferentialParameterSweep object wsing the keyword `num_diff_samples`
            e.g.

        .. highlight::

            {"fs.a": {"diff_mode": "sum", \
                      "diff_sample_type": NormalSample, \
                      "std_dev": 0.01, \
                      "pyomo_object": m.fs.input["a"]}, \
             "fs.b": {"diff_mode": "product", \
                      "diff_sample_type": UniformSample, \
                      "relative_lb": 0.01, \
                      "relative_ub": 0.01, \
                      "pyomo_object": m.fs.input["b"]}, \
             "fs.c": {"diff_mode": "sum", \
                      "diff_sample_type": GeomSample, \
                      "relative_lb": 0.01, \
                      "relative_ub": 10.0, \
                      "pyomo_object": m.fs.input["c"]}} \

        outputs (optional) : An optional dictionary containing "short names" as keys and and Pyomo objects
                  on ``model`` whose values to report as values. E.g.,
                  ``outputs['Short/Pretty-print Name'] = model.fs.variable_or_expression_to_report``.
                  If not provided, i.e., outputs = None, the default behavior is to save all model
                  variables, parameters, and expressions which provides very thorough results
                  at the cost of large file sizes.

        csv_results_file_name (optional) : The path and file name to write a csv file. The default `None`
                                           does not write a csv file.

        h5_results_file_name (optional) : The path and file name to write a h5 file. The default `None`
                                          does not write a file.
                                          Writing an h5 file will also create a companion text file `{h5_results_file_name}.txt`
                                          which contains the variable names contained within the H5 file.

        h5_parent_group_name (optional) : Parent h5 groups for the parameter sweep inputs and outputs to be embedded in.
                                          The default is `None` and it accepts a string for the h5 group.

        optimize_function (optional) : A user-defined function to perform the optimization of flowsheet
                                       ``model`` and loads the results back into ``model``. The first
                                       argument of this function is ``model``. The default uses the
                                       default IDAES solver, raising an exception if the termination
                                       condition is not optimal.

        optimize_kwargs (optional) : Dictionary of kwargs to pass into every call to
                                     ``optimize_function``. The first arg will always be ``model``,
                                     e.g., ``optimize_function(model, **optimize_kwargs)``. The default
                                     uses no kwargs.

        reinitialize_function (optional) : A user-defined function to perform the re-initialize the
                                           flowsheet ``model`` if the first call to ``optimize_function``
                                           fails for any reason. After ``reinitialize_function``, the
                                           parameter sweep tool will immediately call
                                           ``optimize_function`` again.

        reinitialize_kwargs (optional) : Dictionary or kwargs to pass into every call to
                                         ``reinitialize_function``. The first arg will always be
                                         ``model``, e.g.,
                                         ``reinitialize_function(model, **reinitialize_kwargs)``.
                                         The default uses no kwargs.

        reinitialize_before_sweep (optional): Boolean option to reinitialize the flow sheet model before
                                              every parameter sweep realization. The default is False.
                                              Note the parameter sweep model will try to reinitialize the
                                              solve regardless of the option if the run fails.

        probe_function (optional): A user-defined function that can cheaply check if a current model
                                  configuration is solvable without actually reinitializing or solving.

        debugging_data_dir (optional) : Save results on a per-process basis for parallel debugging
                                        purposes. If None no `debugging` data will be saved.

        interpolate_nan_outputs (optional) : When the parameter sweep has finished, interior values
                                             of np.nan will be replaced with a value obtained via
                                             a linear interpolation of their surrounding valid neighbors.
                                             If true, a second output file with the extension "_clean"
                                             will be saved alongside the raw (un-interpolated) values.

        num_samples (optional) : If the user is using sampling techniques rather than a grid
                                 of values, they need to set the required number of samples. This is the
                                 number of base solves that the user desires. The differential sweep will
                                 be performed on these base solves.

        num_diff_samples (optional) : Number of samples for the differential sweep at a given base value.

        seed (optional) : If the user is using a random sampling technique, this sets the seed

    Returns:

        save_data : A list were the first N columns are the values of the parameters passed
                    by ``sweep_params`` and the remaining columns are the values of the
                    simulation identified by the ``outputs`` argument.
    """

    kwargs = {}
    kwargs["build_differential_sweep_specs"] = build_differential_sweep_specs
    if csv_results_file_name is not None:
        kwargs["csv_results_file_name"] = csv_results_file_name
    if h5_results_file_name is not None:
        kwargs["h5_results_file_name"] = h5_results_file_name
    if h5_parent_group_name is not None:
        kwargs["h5_parent_group_name"] = h5_parent_group_name
    if optimize_function is not None:
        kwargs["optimize_function"] = optimize_function
    if optimize_kwargs is not None:
        kwargs["optimize_kwargs"] = optimize_kwargs
    if initialize_function is not None:
        kwargs["initialize_function"] = initialize_function
    if initialize_kwargs is not None:
        kwargs["initialize_kwargs"] = initialize_kwargs
    kwargs["initialize_before_sweep"] = initialize_before_sweep
    if probe_function is not None:
        kwargs["probe_function"] = probe_function
    if debugging_data_dir is not None:
        kwargs["debugging_data_dir"] = debugging_data_dir
    if interpolate_nan_outputs is not None:
        kwargs["interpolate_nan_outputs"] = interpolate_nan_outputs
    kwargs["num_diff_samples"] = num_diff_samples
    kwargs["guarantee_solves"] = guarantee_solves

    dps = DifferentialParameterSweep(**kwargs)

    return dps.parameter_sweep(
        build_model,
        build_sweep_params,
        build_outputs=build_outputs,
        seed=seed,
        num_samples=num_samples,
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import units as pyunits, ExternalFunction
from idaes.core.util.functions import functions_lib


def delta_temperature_chen_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD) approximation given by Chen (1987).  It can be
    supplied to "delta_temperature_callback" HeatExchanger configuration option.
    This uses a cube root function that works with negative numbers returning
    the real negative root. This should always evaluate successfully.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out
    temp_units = pyunits.get_units(dT1)

    # external function that ruturns the real root, for the cuberoot of negitive
    # numbers, so it will return without error for positive and negitive dT.
    b.cbrt = ExternalFunction(
        library=functions_lib(), function="cbrt", arg_units=[temp_units**3]
    )

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return b.cbrt(dT1[t] * dT2[t] * 0.5 * (dT1[t] + dT2[t])) * temp_units



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for ASM1 reaction package.

Authors: Andrew Lee
"""
import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    units,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.models.unit_models import CSTR
from idaes.core import MaterialFlowBasis
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom

from watertap.property_models.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)
from watertap.property_models.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
    ASM1ReactionBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = ASM1ParameterBlock()
        model.rparams = ASM1ReactionParameterBlock(property_package=model.pparams)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rparams.reaction_block_class is ASM1ReactionBlock

        assert len(model.rparams.rate_reaction_idx) == 8
        for i in model.rparams.rate_reaction_idx:
            assert i in ["R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8"]

        assert len(model.rparams.rate_reaction_stoichiometry) == 8 * 14
        for i in model.rparams.rate_reaction_stoichiometry:
            assert i[0] in ["R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8"]
            assert i[1] == "Liq"
            assert i[2] in [
                "H2O",
                "S_I",
                "S_S",
                "X_I",
                "X_S",
                "X_BH",
                "X_BA",
                "X_P",
                "S_O",
                "S_NO",
                "S_NH",
                "S_ND",
                "X_ND",
                "S_ALK",
            ]

        assert isinstance(model.rparams.Y_A, Var)
        assert value(model.rparams.Y_A) == 0.24
        assert isinstance(model.rparams.Y_H, Var)
        assert value(model.rparams.Y_H) == 0.67
        assert isinstance(model.rparams.f_p, Var)
        assert value(model.rparams.f_p) == 0.08
        assert isinstance(model.rparams.i_xb, Var)
        assert value(model.rparams.i_xb) == 0.08
        assert isinstance(model.rparams.i_xp, Var)
        assert value(model.rparams.i_xp) == 0.06

        assert isinstance(model.rparams.mu_A, Var)
        assert value(model.rparams.mu_A) == 0.5
        assert isinstance(model.rparams.mu_H, Var)
        assert value(model.rparams.mu_H) == 4
        assert isinstance(model.rparams.K_S, Var)
        assert value(model.rparams.K_S) == 10e-3
        assert isinstance(model.rparams.K_OH, Var)
        assert value(model.rparams.K_OH) == 0.2e-3
        assert isinstance(model.rparams.K_OA, Var)
        assert value(model.rparams.K_OA) == 0.4e-3
        assert isinstance(model.rparams.K_NO, Var)
        assert value(model.rparams.K_NO) == 0.5e-3
        assert isinstance(model.rparams.b_H, Var)
        assert value(model.rparams.b_H) == 0.3
        assert isinstance(model.rparams.b_A, Var)
        assert value(model.rparams.b_A) == 0.05
        assert isinstance(model.rparams.eta_g, Var)
        assert value(model.rparams.eta_g) == 0.8
        assert isinstance(model.rparams.eta_h, Var)
        assert value(model.rparams.eta_h) == 0.8
        assert isinstance(model.rparams.k_h, Var)
        assert value(model.rparams.k_h) == 3
        assert isinstance(model.rparams.K_X, Var)
        assert value(model.rparams.K_X) == 0.1
        assert isinstance(model.rparams.K_NH, Var)
        assert value(model.rparams.K_NH) == 1e-3
        assert isinstance(model.rparams.k_a, Var)
        assert value(model.rparams.k_a) == 50


class TestReactionBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = ASM1ParameterBlock()
        model.rparams = ASM1ReactionParameterBlock(property_package=model.pparams)

        model.props = model.pparams.build_state_block([1])

        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rxns[1].conc_mass_comp_ref is model.props[1].conc_mass_comp

    @pytest.mark.unit
    def test_rxn_rate(self, model):
        assert isinstance(model.rxns[1].reaction_rate, Var)
        assert len(model.rxns[1].reaction_rate) == 8
        assert isinstance(model.rxns[1].rate_expression, Constraint)
        assert len(model.rxns[1].rate_expression) == 8

    @pytest.mark.unit
    def test_get_reaction_rate_basis(self, model):
        assert model.rxns[1].get_reaction_rate_basis() == MaterialFlowBasis.mass

    @pytest.mark.component
    def test_initialize(self, model):
        assert model.rxns.initialize() is None

    @pytest.mark.unit
    def check_units(self, model):
        assert_units_consistent(model)


class TestReactor:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ASM1ParameterBlock()
        m.fs.rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props)

        m.fs.R1 = CSTR(property_package=m.fs.props, reaction_package=m.fs.rxn_props)

        # Feed conditions based on manual mass balance of inlet and recycle streams
        m.fs.R1.inlet.flow_vol.fix(92230 * units.m**3 / units.day)
        m.fs.R1.inlet.temperature.fix(298.15 * units.K)
        m.fs.R1.inlet.pressure.fix(1 * units.atm)
        m.fs.R1.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "S_S"].fix(14.6112 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_I"].fix(1149 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_S"].fix(89.324 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_BH"].fix(2542.03 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_BA"].fix(148.6 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_P"].fix(448 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "S_O"].fix(0.3928 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "S_NO"].fix(8.32 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "S_NH"].fix(7.696 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "S_ND"].fix(1.9404 * units.g / units.m**3)
        m.fs.R1.inlet.conc_mass_comp[0, "X_ND"].fix(5.616 * units.g / units.m**3)
        m.fs.R1.inlet.alkalinity.fix(4.704 * units.mol / units.m**3)

        m.fs.R1.volume.fix(1000 * units.m**3)

        return m

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_unit_consistency(self, model):
        assert_units_consistent(model) == 0

    @pytest.mark.component
    def test_solve(self, model):
        model.fs.R1.initialize()

        solver = get_solver()
        results = solver.solve(model, tee=True)
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        assert value(model.fs.R1.outlet.flow_vol[0]) == pytest.approx(1.0675, rel=1e-4)
        assert value(model.fs.R1.outlet.temperature[0]) == pytest.approx(
            298.15, rel=1e-4
        )
        assert value(model.fs.R1.outlet.pressure[0]) == pytest.approx(101325, rel=1e-4)
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_I"]) == pytest.approx(
            30e-3, rel=1e-5
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_S"]) == pytest.approx(
            2.81e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_I"]) == pytest.approx(
            1149e-3, rel=1e-3
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_S"]) == pytest.approx(
            82.1e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_BH"]) == pytest.approx(
            2552e-3, rel=1e-3
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_BA"]) == pytest.approx(
            149e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_P"]) == pytest.approx(
            449e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_O"]) == pytest.approx(
            4.3e-6, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_NO"]) == pytest.approx(
            5.36e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_NH"]) == pytest.approx(
            7.92e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "S_ND"]) == pytest.approx(
            1.22e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.conc_mass_comp[0, "X_ND"]) == pytest.approx(
            5.29e-3, rel=1e-2
        )
        assert value(model.fs.R1.outlet.alkalinity[0]) == pytest.approx(
            4.93e-3, rel=1e-2
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a coagulation/flocculation unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_pt, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("CoagulationFlocculationZO")
class CoagulationFlocculationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Coagulation/Flocculation unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "coag_and_floc"

        build_pt(self)

        self.alum_dose = Var(
            self.flowsheet().time,
            units=pyunits.mg / pyunits.L,
            bounds=(0, None),
            doc="Dosing rate of alum",
        )

        self.polymer_dose = Var(
            self.flowsheet().time,
            units=pyunits.mg / pyunits.L,
            bounds=(0, None),
            doc="Dosing rate of polymer",
        )

        self.anion_to_cation_polymer_ratio = Var(
            self.flowsheet().time,
            bounds=(0, None),
            units=pyunits.dimensionless,
            doc="Ratio of anionic to cationic polymer in dosage",
        )

        self.anionic_polymer_dose = Var(
            self.flowsheet().config.time,
            bounds=(0, None),
            units=pyunits.mg / pyunits.L,
            doc="Dosing rate of anionic polymer",
        )

        self.cationic_polymer_dose = Var(
            self.flowsheet().config.time,
            bounds=(0, None),
            units=pyunits.mg / pyunits.L,
            doc="Dosing rate of cationic polymer",
        )

        self.chemical_flow_mass = Var(
            self.flowsheet().time,
            ["alum", "polymer"],
            units=pyunits.kg / pyunits.s,
            bounds=(0, None),
            doc="Mass flow rate of chemical solution",
        )

        self.rapid_mix_retention_time = Var(
            self.flowsheet().config.time,
            units=pyunits.seconds,
            doc="Rapid Mix Retention Time",
        )

        self.floc_retention_time = Var(
            self.flowsheet().config.time,
            units=pyunits.minutes,
            doc="Floc Retention Time",
        )

        self.rapid_mix_basin_vol = Var(
            units=pyunits.m**3, doc="Rapid Mix Basin Volume"
        )

        self.floc_basin_vol = Var(units=pyunits.m**3, doc="Floc Basin Volume")

        self.num_rapid_mixers = Var(
            units=pyunits.dimensionless, doc="Number of Rapid Mixers"
        )

        self.num_floc_mixers = Var(
            units=pyunits.dimensionless, doc="Number of Floc Mixers"
        )

        self.num_rapid_mix_processes = Var(
            units=pyunits.dimensionless, doc="Number of Rapid Mix Processes"
        )

        self.num_floc_processes = Var(
            units=pyunits.dimensionless, doc="Number of Floc Processes"
        )

        self.num_coag_processes = Var(
            units=pyunits.dimensionless, doc="Number of Coagulation Processes"
        )

        self.num_floc_injection_processes = Var(
            units=pyunits.dimensionless, doc="Number of Floc Injection Processes"
        )

        self.velocity_gradient_rapid_mix = Var(
            self.flowsheet().config.time,
            units=pyunits.s**-1,
            doc="Rapid Mix Velocity Gradient",
        )

        self.velocity_gradient_floc = Var(
            self.flowsheet().config.time,
            units=pyunits.s**-1,
            doc="Floc Velocity Gradient",
        )

        self.power_rapid_mix = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            doc="Rapid Mix Power Consumption",
        )

        self.power_floc = Var(
            self.flowsheet().config.time, units=pyunits.kW, doc="Floc Power Consumption"
        )

        self.electricity = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            doc="Total Power Consumption",
        )

        self._fixed_perf_vars.append(self.alum_dose)
        self._fixed_perf_vars.append(self.polymer_dose)
        self._fixed_perf_vars.append(self.anion_to_cation_polymer_ratio)
        self._fixed_perf_vars.append(self.rapid_mix_retention_time)
        self._fixed_perf_vars.append(self.floc_retention_time)
        self._fixed_perf_vars.append(self.num_floc_injection_processes)
        self._fixed_perf_vars.append(self.num_floc_processes)
        self._fixed_perf_vars.append(self.num_rapid_mixers)
        self._fixed_perf_vars.append(self.num_coag_processes)
        self._fixed_perf_vars.append(self.num_floc_mixers)
        self._fixed_perf_vars.append(self.num_rapid_mix_processes)
        self._fixed_perf_vars.append(self.velocity_gradient_rapid_mix)
        self._fixed_perf_vars.append(self.velocity_gradient_floc)

        self._perf_var_dict["Alum Dosage (mg/L)"] = self.alum_dose
        self._perf_var_dict["Polymer Dosage (mg/L)"] = self.polymer_dose
        self._perf_var_dict["Alum Flow (kg/s)"] = self.chemical_flow_mass[0, "alum"]
        self._perf_var_dict["Polymer Flow (kg/s)"] = self.chemical_flow_mass[
            0, "polymer"
        ]
        self._perf_var_dict["Rapid Mix Basin Volume (m^3)"] = self.rapid_mix_basin_vol
        self._perf_var_dict["Floc Basin Volume (m^3)"] = self.floc_basin_vol
        self._perf_var_dict[
            "Rapid Mix Retention Time (s)"
        ] = self.rapid_mix_retention_time
        self._perf_var_dict["Floc Retention Time (min)"] = self.floc_retention_time
        self._perf_var_dict[
            "Rapid Mix Velocity Gradient (1/s)"
        ] = self.velocity_gradient_rapid_mix
        self._perf_var_dict[
            "Floc Velocity Gradient (1/s)"
        ] = self.velocity_gradient_floc
        self._perf_var_dict["Rapid Mix Power (kW)"] = self.power_rapid_mix
        self._perf_var_dict["Floc Power (kW)"] = self.power_floc
        self._perf_var_dict["Total Power Consumption (kW)"] = self.electricity

        def rule_rapid_mix_basin_vol(blk):
            return (
                blk.rapid_mix_basin_vol
                == blk.properties[0].flow_vol * blk.rapid_mix_retention_time[0]
            )

        self.rapid_mix_basin_vol_constraint = Constraint(rule=rule_rapid_mix_basin_vol)

        def rule_floc_basin_vol(blk):
            return blk.floc_basin_vol == pyunits.convert(
                blk.properties[0].flow_vol * blk.floc_retention_time[0],
                to_units=pyunits.m**3,
            )

        self.floc_basin_vol_constraint = Constraint(rule=rule_floc_basin_vol)

        def rule_chem_flow(blk, t, j):
            if j == "alum":
                chemical_dosage = blk.alum_dose[t]
            elif j == "polymer":
                chemical_dosage = blk.polymer_dose[t]
            return blk.chemical_flow_mass[t, j] == pyunits.convert(
                chemical_dosage * blk.properties[t].flow_vol,
                to_units=pyunits.kg / pyunits.s,
            )

        self.chemical_flow_constraint = Constraint(
            self.flowsheet().time, ["alum", "polymer"], rule=rule_chem_flow
        )

        def rule_anionic_polymer_dose(blk, t):
            return blk.anionic_polymer_dose[t] == blk.anion_to_cation_polymer_ratio[
                t
            ] * blk.polymer_dose[t] / (blk.anion_to_cation_polymer_ratio[t] + 1)

        self.anionic_polymer_dose_constraint = Constraint(
            self.flowsheet().config.time, rule=rule_anionic_polymer_dose
        )

        def rule_cationic_polymer_dose(blk, t):
            return blk.cationic_polymer_dose[t] == blk.polymer_dose[t] / (
                blk.anion_to_cation_polymer_ratio[t] + 1
            )

        self.cationic_polymer_dose_constraint = Constraint(
            self.flowsheet().config.time, rule=rule_cationic_polymer_dose
        )

        # TODO: WT3 doesn't include pump electricity. Consider adding after case study validation
        # pump_electricity(self, self.chemical_flow_vol)

        @self.Constraint(
            self.flowsheet().time, doc="Constraint for rapid mix power consumption"
        )
        def rule_power_rapid_mix(b, t):
            return b.power_rapid_mix[t] == pyunits.convert(
                b.num_rapid_mixers
                * b.properties[t].visc_d
                * b.rapid_mix_basin_vol
                * b.velocity_gradient_rapid_mix[t] ** 2,
                to_units=pyunits.kW,
            )

        @self.Constraint(
            self.flowsheet().time, doc="Constraint for floc power consumption"
        )
        def rule_power_floc(b, t):
            return b.power_floc[t] == pyunits.convert(
                b.num_floc_mixers
                * b.properties[t].visc_d
                * b.floc_basin_vol
                * b.velocity_gradient_floc[t] ** 2,
                to_units=pyunits.kW,
            )

        @self.Constraint(self.flowsheet().time, doc="Total power consumption")
        def electricity_constraint(b, t):
            return b.electricity[t] == b.power_floc[t] + b.power_rapid_mix[t]

    @property
    def default_costing_method(self):
        return self.cost_coag_and_floc

    @staticmethod
    def cost_coag_and_floc(blk):
        """
        General method for costing coagulation/flocculation processes. Capital cost
        is based on the alum flowrate and the polymer flowrate of the incoming stream.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F, G, H = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "capital_mix_a_parameter",
                "capital_mix_b_parameter",
                "capital_floc_a_parameter",
                "capital_floc_b_parameter",
                "capital_coag_inj_a_parameter",
                "capital_coag_inj_b_parameter",
                "capital_floc_inj_a_parameter",
                "capital_floc_inj_b_parameter",
            ],
        )

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        cost_rapid_mix = (
            A
            * pyo.units.convert(
                blk.unit_model.rapid_mix_basin_vol, to_units=pyo.units.gallons
            )
            + B
        ) * blk.unit_model.num_rapid_mix_processes

        cost_floc = (
            C
            * pyo.units.convert(
                blk.unit_model.floc_basin_vol, to_units=pyo.units.Mgallons
            )
            + D
        ) * blk.unit_model.num_floc_processes

        cost_coag_inj = (
            E
            * pyo.units.convert(
                blk.unit_model.chemical_flow_mass[t0, "alum"],
                to_units=(pyo.units.lb / pyo.units.hr),
            )
            + F
        ) * blk.unit_model.num_coag_processes

        cost_floc_inj = (
            G
            * pyo.units.convert(
                blk.unit_model.chemical_flow_mass[t0, "polymer"],
                to_units=(pyo.units.lb / pyo.units.day),
            )
            + H
        ) * blk.unit_model.num_floc_injection_processes

        expr = (
            pyo.units.convert(
                cost_rapid_mix,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
            + pyo.units.convert(
                cost_floc, to_units=blk.config.flowsheet_costing_block.base_currency
            )
            + pyo.units.convert(
                cost_coag_inj, to_units=blk.config.flowsheet_costing_block.base_currency
            )
            + pyo.units.convert(
                cost_floc_inj, to_units=blk.config.flowsheet_costing_block.base_currency
            )
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.ui.fsapi import FlowsheetInterface
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.groundwater_treatment.groundwater_treatment import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
)
from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent


def export_to_ui():
    return FlowsheetInterface(
        name="Groundwater treatment",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.feed.flow_vol[0],
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Inlet volumetric flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "arsenic"],
        name="Arsenic concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet arsenic concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "uranium"],
        name="Uranium concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet uranium concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "nitrate"],
        name="Nitrate concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet nitrate concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "phosphates"],
        name="Phosphates concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet phosphates concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "iron"],
        name="Iron concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet iron concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.conc_mass_comp[0, "filtration_media"],
        name="Filtration media concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=2,
        description="Inlet filtration media concentration",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, pump
    exports.add(
        obj=fs.pump.lift_height,
        name="Lift height",
        ui_units=pyunits.m,
        display_units="m",
        rounding=2,
        description="Lift height for pump",
        is_input=True,
        input_category="Pump",
        is_output=False,
    )
    exports.add(
        obj=fs.pump.eta_pump,
        name="Pump efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Efficiency of pump",
        is_input=True,
        input_category="Pump",
        is_output=False,
    )
    exports.add(
        obj=fs.pump.eta_motor,
        name="Motor efficiency",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Efficiency of motor",
        is_input=True,
        input_category="Pump",
        is_output=False,
    )

    # Unit cost data, pump
    exports.add(
        obj=fs.costing.pump_electricity.pump_cost[None],
        name="Pump cost",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.hr),
        display_units="$/(m^3/hr)",
        rounding=0,
        description="Pump capital cost parameter",
        is_input=True,
        input_category="Pump costing",
        is_output=False,
    )

    # Unit model data, microbial battery
    exports.add(
        obj=fs.micbatt.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.reaction_conversion[0, "iron_precipitation"],
        name="Iron conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Iron conversion [g-Fe reacted/g-Fe inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.removal_frac_mass_comp[0, "arsenic"],
        name="Arsenic removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Arsenic removal [g-As removed/g-As inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.removal_frac_mass_comp[0, "uranium"],
        name="Uranium removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Uranium removal [g-U removed/g-U inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.removal_frac_mass_comp[0, "nitrate"],
        name="Nitrate removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Nitrate removal [g-nitrate removed/g-nitrate inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.removal_frac_mass_comp[0, "phosphates"],
        name="Phosphates removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Phosphates removal [g-P removed/g-P inlet]",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.energy_electric_flow_vol_inlet,
        name="Electricity intensity",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="kWh/m3",
        rounding=2,
        description="Electricity intensity relating the intensity to the product volume",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    exports.add(
        obj=fs.micbatt.HRT,
        name="HRT",
        ui_units=pyunits.hr,
        display_units="h",
        rounding=1,
        description="Hydraulic retention time",
        is_input=True,
        input_category="Microbial battery",
        is_output=False,
    )
    # Unit cost data, microbial battery
    exports.add(
        obj=fs.costing.microbial_battery.sizing_cost[None],
        name="Microbial battery cost",
        ui_units=(fs.costing.base_currency * pyunits.day) / pyunits.m**3,
        display_units="($ * day)/m3 of microbial battery",
        rounding=0,
        description="Microbial battery capital cost parameter",
        is_input=True,
        input_category="Microbial battery costing",
        is_output=False,
    )
    # System costing
    exports.add(
        obj=fs.costing.utilization_factor,
        name="Utilization factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Utilization factor - [annual use hours/total hours in year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.TIC,
        name="Practical investment factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=1,
        description="Practical investment factor - [total investment cost/direct "
        "capital costs]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.plant_lifetime,
        name="Plant lifetime",
        ui_units=pyunits.year,
        display_units="years",
        rounding=1,
        description="Plant lifetime",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.wacc,
        name="Discount rate",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Discount rate used in calculating the capital annualization",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.maintenance_costs_percent_FCI,
        name="Fixed operating cost factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Fixed operating cost factor - [annual fixed operating cost/total "
        "investment cost]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kWh",
        rounding=3,
        description="Electricity cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.filtration_media_cost,
        name="Filtration media cost",
        ui_units=fs.costing.base_currency / pyunits.kg,
        display_units="$/kg",
        rounding=3,
        description="Filtration media cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.filtration_media_disposal_cost,
        name="Filtration media disposal cost",
        ui_units=fs.costing.base_currency / pyunits.kg,
        display_units="$/kg",
        rounding=3,
        description="Filtration media disposal cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    # Outlets
    exports.add(
        obj=fs.filtered_water.properties[0].flow_vol,
        name="Filtered water flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet filtered water flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["arsenic"],
        name="Filtered water arsenic concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=3,
        description="Outlet filtered water arsenic concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["uranium"],
        name="Filtered water uranium concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=3,
        description="Outlet filtered water uranium concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["nitrate"],
        name="Filtered water nitrate concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=3,
        description="Outlet filtered water nitrate concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.filtered_water.properties[0].conc_mass_comp["phosphates"],
        name="Filtered water phosphates concentration",
        ui_units=pyunits.mg / pyunits.L,
        display_units="mg/L",
        rounding=3,
        description="Outlet filtered water phosphates concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.byproduct.properties[0].flow_mass_comp["arsenic"],
        name="Byproduct arsenic flow rate",
        ui_units=pyunits.mg / pyunits.day,
        display_units="mg-arsenic/day",
        rounding=5,
        description="Outlet byproduct arsenic flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.byproduct.properties[0].flow_mass_comp["uranium"],
        name="Byproduct uranium flow rate",
        ui_units=pyunits.mg / pyunits.day,
        display_units="mg-uranium/day",
        rounding=5,
        description="Outlet byproduct uranium flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.byproduct.properties[0].flow_mass_comp["nitrate"],
        name="Byproduct nitrate flow rate",
        ui_units=pyunits.mg / pyunits.day,
        display_units="mg-nitrate/day",
        rounding=5,
        description="Outlet byproduct nitrate flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.byproduct.properties[0].flow_mass_comp["phosphates"],
        name="Byproduct phosphates flow rate",
        ui_units=pyunits.mg / pyunits.day,
        display_units="mg-phosphates/day",
        rounding=5,
        description="Outlet byproduct phosphates flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.byproduct.properties[0].flow_mass_comp["filtration_media"],
        name="Byproduct filtration media flow rate",
        ui_units=pyunits.mg / pyunits.day,
        display_units="mg-media/day",
        rounding=5,
        description="Outlet byproduct filtration media flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    # System metrics
    exports.add(
        obj=fs.costing.LCOT,
        name="LCOT",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=3,
        description="Levelized cost of treatment including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LCOW,
        name="LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of treated water",
        rounding=3,
        description="Levelized cost of water including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    # Normalized metrics
    total_capital_norm = fs.costing.total_capital_cost / fs.feed.properties[0].flow_vol
    exports.add(
        obj=total_capital_norm,
        name="Total capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized total capital costs accounting for indirect "
        "capital and installation - [total capital costs/feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    direct_capital_norm = (
        (fs.pump.costing.capital_cost + fs.micbatt.costing.capital_cost)
        / fs.costing.TIC
        / fs.feed.properties[0].flow_vol
    )
    exports.add(
        obj=direct_capital_norm,
        name="Direct capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized direct capital costs - [total direct capital "
        "costs/feed flow rate] ",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    elec_operating_norm = (
        fs.costing.aggregate_flow_costs["electricity"] / fs.costing.annual_water_inlet
    )
    exports.add(
        obj=elec_operating_norm,
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Normalized electricity cost - [annual electricity costs/annual "
        "feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )

    # performance metrics
    recovery_vol = (
        fs.filtered_water.properties[0].flow_vol / fs.feed.properties[0].flow_vol
    )
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="m3 of product/m3 of feed",
        rounding=3,
        description="Normalized volumetric recovery",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_arsenic = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["arsenic"]
        / fs.feed.properties[0].flow_mass_comp["arsenic"]
    )
    exports.add(
        obj=removal_arsenic,
        name="Arsenic removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Arsenic removal fraction [1 - outlet As flow/inlet As flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_uranium = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["uranium"]
        / fs.feed.properties[0].flow_mass_comp["uranium"]
    )
    exports.add(
        obj=removal_uranium,
        name="Uranium removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Uranium removal fraction [1 - outlet U flow/inlet U flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_nitrate = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["nitrate"]
        / fs.feed.properties[0].flow_mass_comp["nitrate"]
    )
    exports.add(
        obj=removal_nitrate,
        name="Nitrate removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Nitrate removal fraction [1 - outlet NO3 flow/inlet NO3 flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_phosphates = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["phosphates"]
        / fs.feed.properties[0].flow_mass_comp["phosphates"]
    )
    exports.add(
        obj=removal_phosphates,
        name="Phosphates removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Phosphates removal fraction [1 - outlet P flow/inlet P flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_iron = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["iron"]
        / fs.feed.properties[0].flow_mass_comp["iron"]
    )
    exports.add(
        obj=removal_iron,
        name="Iron removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Iron removal fraction [1 - outlet Fe flow/inlet Fe flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_filtration_media = (
        1
        - fs.filtered_water.properties[0].flow_mass_comp["filtration_media"]
        / fs.feed.properties[0].flow_mass_comp["filtration_media"]
    )
    exports.add(
        obj=removal_filtration_media,
        name="Filtration media removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Filtration media removal fraction [1 - outlet media flow/inlet media flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    # Capital costs
    exports.add(
        obj=fs.costing.total_capital_cost,
        name="Total",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Total capital costs - including investment factor to account "
        "for indirect capital",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.pump.costing.capital_cost,
        name="Pump",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Pump",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.micbatt.costing.capital_cost,
        name="Microbial battery",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Microbial battery",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    # Operating costs
    exports.add(
        obj=fs.costing.total_operating_cost,
        name="Total",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Total annual operating costs - including electricity, heating, "
        "and fixed",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["electricity"],
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual electricity costs ",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.total_fixed_operating_cost,
        name="Fixed",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual fixed operating costs - these costs include material "
        "replacement, maintenance, and labor",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    add_costing(m)
    assert_degrees_of_freedom(m, 0)
    m.fs.costing.initialize()

    results = solve(m)
    assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import cost_membrane, register_costing_parameter_block


class OAROType(StrEnum):
    standard = "standard"
    high_pressure = "high_pressure"


def build_osmotically_assisted_reverse_osmosis_cost_param_block(blk):
    """
    Reference parameter values are provided from:
    Bartholomew, T. V., Mey, L., Arena, J. T., Siefert, N. S., & Mauter, M. S. (2017).
    Osmotically assisted reverse osmosis for high salinity brine treatment. Desalination, 421, 3-11.
    """
    blk.factor_membrane_replacement = pyo.Var(
        initialize=0.15,
        doc="Membrane replacement factor [fraction of membrane replaced/year]",
        units=pyo.units.year**-1,
    )
    blk.membrane_cost = pyo.Var(
        initialize=30,
        doc="Membrane cost",
        units=pyo.units.USD_2018 / (pyo.units.meter**2),
    )
    blk.high_pressure_membrane_cost = pyo.Var(
        initialize=50,
        doc="Membrane cost",
        units=pyo.units.USD_2018 / (pyo.units.meter**2),
    )


@register_costing_parameter_block(
    build_rule=build_osmotically_assisted_reverse_osmosis_cost_param_block,
    parameter_block_name="osmotically_assisted_reverse_osmosis",
)
def cost_osmotically_assisted_reverse_osmosis(blk, oaro_type=OAROType.standard):
    """
    Osmotically assisted reverse osmosis costing method

    TODO: describe equations

    Args:
        oaro_type: ROType Enum indicating osmotically assisted reverse osmosis type,
            default = OAROType.standard
    """
    if oaro_type == OAROType.standard:
        return cost_membrane(
            blk,
            blk.costing_package.osmotically_assisted_reverse_osmosis.membrane_cost,
            blk.costing_package.osmotically_assisted_reverse_osmosis.factor_membrane_replacement,
        )
    elif oaro_type == OAROType.high_pressure:
        return cost_high_pressure_osmotically_assisted_reverse_osmosis(blk)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for oaro_type:"
            f" {oaro_type}. Argument must be a member of the OAROType Enum."
        )


@register_costing_parameter_block(
    build_rule=build_osmotically_assisted_reverse_osmosis_cost_param_block,
    parameter_block_name="osmotically_assisted_reverse_osmosis",
)
def cost_high_pressure_osmotically_assisted_reverse_osmosis(blk):
    cost_membrane(
        blk,
        blk.costing_package.osmotically_assisted_reverse_osmosis.high_pressure_membrane_cost,
        blk.costing_package.osmotically_assisted_reverse_osmosis.factor_membrane_replacement,
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
    Simple unit tests for example flowsheet of SepRO with Chlorination.

    NOTE: That flowsheet is not meant to be viewed as a final product, but
    a sample of how to incorporate more complex chemistry into a simple
    flowsheet.
"""
import pytest

from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.chemistry.posttreatment_ideal_naocl_chlorination_block import (
    run_ideal_naocl_mixer_example,
    run_ideal_naocl_chlorination_example,
    run_chlorination_block_example,
    build_translator_from_RO_to_chlorination_block,
    unfix_ideal_naocl_mixer_inlet_stream,
)
from watertap.examples.flowsheets.full_treatment_train.model_components import (
    property_models,
)
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    desalination,
)
from watertap.examples.flowsheets.full_treatment_train.util import check_dof
from pyomo.environ import TransformationFactory
from pyomo.network import Arc

__author__ = "Austin Ladshaw, Xinhong Liu"


@pytest.mark.component
def test_ideal_naocl_mixer():
    model = run_ideal_naocl_mixer_example()
    assert model.fs.ideal_naocl_mixer_unit.dosing_rate.value == pytest.approx(
        0.4e-6, rel=1e-3
    )
    assert model.fs.ideal_naocl_mixer_unit.outlet.flow_mol[0].value == pytest.approx(
        10.00001074690966, rel=1e-3
    )
    assert model.fs.ideal_naocl_mixer_unit.outlet.mole_frac_comp[
        0, "OCl_-"
    ].value == pytest.approx(5.373448801531194e-07, rel=1e-3)


@pytest.mark.component
def test_ideal_naocl_chlorination():
    model = run_ideal_naocl_chlorination_example()
    assert model.fs.ideal_naocl_chlorination_unit.free_chlorine.value == pytest.approx(
        1.893849296278168, rel=1e-3
    )
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[
        0, "OCl_-"
    ].value == pytest.approx(4.254858076511e-07, rel=1e-3)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[
        0, "H_+"
    ].value == pytest.approx(5.64e-11, rel=1e-1)


@pytest.mark.component
def test_ideal_naocl_chlorination_full_block():
    model = run_chlorination_block_example(fix_free_chlorine=True)
    assert model.fs.ideal_naocl_mixer_unit.dosing_rate.value == pytest.approx(
        9.505698559578499e-07, rel=1e-3
    )
    assert model.fs.ideal_naocl_mixer_unit.outlet.flow_mol[0].value == pytest.approx(
        25.000025535888078, rel=1e-3
    )
    assert model.fs.ideal_naocl_mixer_unit.outlet.mole_frac_comp[
        0, "OCl_-"
    ].value == pytest.approx(5.107133802822922e-07, rel=1e-3)

    assert model.fs.ideal_naocl_chlorination_unit.free_chlorine.value == pytest.approx(
        2, rel=1e-3
    )
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[
        0, "OCl_-"
    ].value == pytest.approx(4.48e-07, rel=1e-2)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[
        0, "H_+"
    ].value == pytest.approx(5.43e-11, rel=1e-1)


@pytest.mark.component
def test_addition_of_translator():
    model = run_ideal_naocl_mixer_example()
    property_models.build_prop(model, base="TDS")
    build_translator_from_RO_to_chlorination_block(model)
    assert hasattr(model.fs, "RO_to_Chlor")


@pytest.mark.component
def test_build_flowsheet():
    model = run_chlorination_block_example(fix_free_chlorine=True)
    property_models.build_prop(model, base="TDS")
    build_translator_from_RO_to_chlorination_block(model)
    # Here, we set 'has_feed' to True because RO is our first block in the flowsheet
    kwargs_desal = {
        "has_desal_feed": True,
        "is_twostage": False,
        "has_ERD": False,
        "RO_type": "0D",
        "RO_base": "TDS",
        "RO_level": "detailed",
    }
    desal_port = desalination.build_desalination(model, **kwargs_desal)
    desalination.scale_desalination(model, **kwargs_desal)
    desalination.initialize_desalination(model, **kwargs_desal)

    # Add the connecting arcs
    model.fs.S1 = Arc(source=desal_port["out"], destination=model.fs.RO_to_Chlor.inlet)
    model.fs.S2 = Arc(
        source=model.fs.RO_to_Chlor.outlet,
        destination=model.fs.ideal_naocl_mixer_unit.inlet_stream,
    )
    TransformationFactory("network.expand_arcs").apply_to(model)

    # Unfix inlet stream for mixer
    unfix_ideal_naocl_mixer_inlet_stream(model)
    check_dof(model)
